#!/usr/local/bin/ruby -w
require 'rconftool'
require 'test/unit'
require 'fileutils'

# Test cases for rconftool, done in a messy way. But messy tests are
# better than no tests!

class RconftoolTest < Test::Unit::TestCase
  FN1 = "_test_file.dist"
  FN2 = "_test_file"
  FN3 = "_test_file.old"

  def teardown
    File.delete(FN1) rescue nil
    File.delete(FN2) rescue nil
    File.delete(FN2+".bak") rescue nil
    File.delete(FN3) rescue nil
  end

  def doit(distdata, targetdata, finaldata)
    File.open(FN1,"w") { |f| f.write(distdata) }
    File.open(FN2,"w") { |f| f.write(targetdata) }
    Rconftool::install(FN1,FN2)
    result = nil
    File.open(FN2) { |f| result = f.read }
    assert_equal(finaldata, result)
  end

  # If target file does not exist, it's a verbatim copy
  def test_notarget
    data = <<ENDA
##VERSION:123
# This is a basic test
##NAME: FOO:0
FOOBAR=X
ENDA
    File.open(FN1,"w") { |f| f.write(data) }
    File.delete(FN2) rescue nil
    Rconftool::install(FN1,FN2)
    result = nil
    File.open(FN2) { |f| result = f.read }
    assert_equal(data, result)
  end

  # If VERSION header is the same, the target is untouched
  def test_sameversion
    doit(<<ENDA,<<ENDB,<<ENDC)
# Random
##VERSION:123
##NAME: FOO:0
FOOBAR=X
ENDA
# Other
# Stuff
##VERSION:123
##NAME: BAR:1
BAZ=Y
ENDB
# Other
# Stuff
##VERSION:123
##NAME: BAR:1
BAZ=Y
ENDC
  end

  # If src and dest VERSIONS are different, a merge is done
  def test_merge
    doit(<<ENDA,<<ENDB,<<ENDC)
# Random
##VERSION:123
##NAME: FOO:0
# This item is carried forward
FOOBAR=X
##NAME: BAR:0
# This item does not exist in the destfile so is copied verbatim
WIBBLE=ABC
##NAME: BAZ:1
# Version is different in destfile, so we override
SOMENEWSETTING=123
OTHERNEWSETTING=456
ENDA
# Flurble
##VERSION:99
##NAME: FOO:0
# Old comment is dropped but old value carried forward

FOOBAR=YYY

##NAME: ZOG:0
# This item does not exist in the srcfile so is dropped
ZEBEDEE=BOING
##NAME: BAZ:0
# This old value is replaced, but kept as a comment in the new file
OLDSETTING=9
SOMENEWSETTING=99
ENDB
# Random
##VERSION:123
##NAME: FOO:0
# This item is carried forward
#
# DEFAULT SETTING from #{FN1}:
#FOOBAR=X

FOOBAR=YYY

##NAME: BAR:0
# This item does not exist in the destfile so is copied verbatim
WIBBLE=ABC
##NAME: BAZ:1
# Version is different in destfile, so we override
#
# Previous setting (inserted by rconftool):
#
#OLDSETTING=9
#SOMENEWSETTING=99
SOMENEWSETTING=123
OTHERNEWSETTING=456
ENDC
  end

  # If src has VERSION and target does not have one, the target is
  # overwritten
  def test_newversion
    doit(<<ENDA,<<ENDB,<<ENDC)
# Random
##VERSION:123
##NAME: FOO:0
FOOBAR=X
ENDA
# This is a file
# without a VERSION string
FOOBAR=Y
ENDB
# Random
##VERSION:123
##NAME: FOO:0
FOOBAR=X
ENDC
  end

  # Test we get an exception if the source file does not exist, and what
  # happens if the source does not contain a VERSION string
  def test_exists
    assert_raises(Errno::ENOENT) {
      Rconftool::install('_nonexistent.dist','_nonexistent')
    }
    File.open(FN1,"w") { |f| f.write("#wibble\n") }
    File.delete(FN2) rescue nil
    #Old behaviour:
    #assert_raises(Rconftool::NoVersionLine) {
    #  Rconftool::install(FN1,FN2)
    #}
    #Now we just install the file verbatim
    Rconftool::install(FN1,FN2)
    result = nil
    File.open(FN2) { |f| result = f.read }
    assert_equal("#wibble\n", result)
    #Try it again, when the file already exists; it should not be touched
    File.open(FN1,"w") { |f| f.write("#boing\n") }
    Rconftool::install(FN1,FN2)
    File.open(FN2) { |f| result = f.read }
    assert_equal("#wibble\n", result)
    #Try it again, with a sysconftool format file; it should overwrite
    File.open(FN1,"w") { |f| f.write("##VERSION:123\n") }
    Rconftool::install(FN1,FN2)
    File.open(FN2) { |f| result = f.read }
    assert_equal("##VERSION:123\n", result)
  end

  # Test that the target file is chmod'd to the same as the source
  def test_chmod
    File.open(FN1,"w") { |f| f.write("##VERSION:123\n") }
    File.chmod(0614, FN1)
    File.open(FN2,"w") { |f| f.write("##VERSION:456\n") }
    File.chmod(0444, FN2)
    Rconftool::install(FN1,FN2)
    statbuf = File.stat(FN2)
    assert_equal(0614, statbuf.mode & 0777)  # This is a Unix-ism
  end

  # Check the three-way version of Rconftool::install. This allows
  # merged-in values to be picked up from a different location if the
  # target file does not already exist
  def test_threeway
    File.open(FN1,"w") { |f| f.write(<<ENDA) }
##VERSION:123
# This is the distfile
##NAME:FOO:0
FOO=1
ENDA
    File.delete(FN2) rescue nil
    File.open(FN3,"w") { |f| f.write(<<ENDB) }
##VERSION:456
# This is the old config file
##NAME:FOO:0
FOO=2
ENDB
    Rconftool::install(FN1,FN2,FN3)
    result = nil
    File.open(FN2) { |f| result = f.read }
    assert_equal(<<ENDC, result)
##VERSION:123
# This is the distfile
##NAME: FOO:0
#
# DEFAULT SETTING from #{FN1}:
#FOO=1
FOO=2
ENDC
  end

  # Test capturing debug messages into a string
  def test_string
    debug = ""
    File.open(FN1,"w") { |f| f.write(<<ENDA) }
##VERSION:1
##NAME: FOO:0
ENDA
    File.open(FN2,"w") { |f| f.write(<<ENDB) }
##VERSION:0
ENDB
    Rconftool::install(FN1,FN2,nil, :debug=>debug)
    assert_equal(<<ENDC, debug)
#{FN2}:
  FOO: new
ENDC
  end

  # Test directory recursion helper function
  TEST_DIR = "_test_dir"
  SUBDIR_LIST = ["_test_subdir1","_test_subdir2","_test_subdir2/xxx"]
  RECUR_LIST = ["file1.dist", "file2.dist",
                "_test_subdir1/file3.dist",
                "_test_subdir2/file4.dist",
		"_test_subdir2/xxx/file5.dist"]
  def test_recurse
    FileUtils.rm_rf(TEST_DIR) rescue nil
    Dir.mkdir(TEST_DIR)
    SUBDIR_LIST.each { |d| Dir.mkdir(TEST_DIR+"/"+d) }
    RECUR_LIST.each do |fn|
      File.open(TEST_DIR+"/"+fn,"w") { |f| f.write("abc") }
    end
    result = []
    Rconftool::recurse_dir(TEST_DIR) { |d| result << d }
    assert_equal((RECUR_LIST+SUBDIR_LIST).sort, result.sort)
    FileUtils.rm_rf(TEST_DIR)
  end

  # Test the command-line interface in simplest form. We rely on the
  # fact that FN1 ends in '.dist' and this is stripped off to get the
  # target filename FN2.
  def test_cmd_simple
    File.open(FN1,"w") { |f| f.write(<<ENDA) }
##VERSION:1
##NAME: FOO:0
X=123
ENDA
    File.open(FN2,"w") { |f| f.write(<<ENDB) }
##VERSION:0
##NAME: FOO:0
Y=456
ENDB
    argv = [FN1]
    p = Rconftool::Processor.new(argv)
    p.run(argv)
    result = nil
    File.open(FN2) { |f| result = f.read }
    assert_equal(<<ENDC, result)
##VERSION:1
##NAME: FOO:0
#
# DEFAULT SETTING from #{FN1}:
#X=123
Y=456
ENDC
  end

  # Test the command-line interface with recursive option, suffix
  # stripping and adding, and making intermediate directories
  TARGET_DIR="_target_dir"
  def test_cmd_complex
    FileUtils.rm_rf(TEST_DIR) rescue nil
    FileUtils.rm_rf(TARGET_DIR) rescue nil
    Dir.mkdir(TEST_DIR)
    SUBDIR_LIST.each { |d| Dir.mkdir(TEST_DIR+"/"+d) }
    File.chmod(0715, TEST_DIR+"/"+SUBDIR_LIST[0])  # test mode bits copied

    RECUR_LIST.each do |fn|
      File.open(TEST_DIR+"/"+fn,"w") { |f| f.write("##VERSION:#{fn}\n") }
    end
    File.open(FN1,"w") { |f| f.write("##VERSION:#{FN1}") }

    Dir.mkdir(TARGET_DIR)
    argv = ["--strip-suffix",".dist","--add-suffix",".conf","--recursive",
            TEST_DIR, FN1, "--targetdir", TARGET_DIR]
    p = Rconftool::Processor.new(argv)
    p.run(argv)

    reslist = Dir["_target_dir/**/*"]
    reslist.delete_if { |d| File.directory?(d) }
    expected = RECUR_LIST.collect { |n| TARGET_DIR+"/"+n.sub(/\.dist$/,'.conf') }
    expected << TARGET_DIR+"/"+FN1.sub(/\.dist$/,'.conf')
    assert_equal(expected.sort, reslist.sort)
    assert_equal(0715, File.stat(TARGET_DIR+"/"+SUBDIR_LIST[0]).mode & 0777)
    FileUtils.rm_rf(TEST_DIR)
    FileUtils.rm_rf(TARGET_DIR)
  end
end
